//  $Id: prom_tabl.hh,v 1.2 2004/06/16 00:12:25 adams Exp $
//  Author:             Mark F. Adams
// Copyright (c) 2000 by Mark F. Adams 
//  Filename:           prom_tabl.hh
//
#ifndef __TABLE_H__
#define __TABLE_H__
// abort() usage further below requires stdlib
#include <stdlib.h>

#define HASH_FACT 79943
#define HASH(x) ((unsigned long)((HASH_FACT*(unsigned long)x)%bits.tablesize))
#define PINT(x)  ((PromStack*)(x)-(PromStack*)(0))
#define HASHINC(hash) ((hash+1)%(bits.tablesize))
#define prom_min(a,b) ( a < b ? a : b )

#if defined(PROM_HAVE_PETSC)
#define PROMNEW(v,sz,type){int i=PetscMalloc(sizeof(type)*sz,&v);CHKERRQ(i);}
#define DELETE(v) { PetscFree(v); } 
#define MEMZERO(v,sz,type) PetscMemzero( (char*)v, sz*sizeof(type) )
#else
#define PROMNEW(v,sz,type) {v = new type[sz]; CHKNEW(v,1);}
#define DELETE(v) {delete [] v; PromContext::numnews_--;}
//#define MEMZERO(v,sz,type){for(int i=0;i<sz;i++)v[i]=(type)0;}
#define MEMZERO(v,sz,type) {memzero( (char*)v, sz*sizeof(type) )}
#endif

typedef int TablePos;
class AthenaNode;
class PromDTet;
class PromNode;
extern int prom_table_size;

/////////////////////////////////////////////////////////////////////////////
// PromTable
//
template <class T> class PromTable 
{
public:
  PromTable( int size = -1 ) :
    keytable(NULL), table(NULL)
  {
    if( size < 0 ) size = prom_table_size;
    Init( size ); 
  }
  // construction
  int Init( int size ){
    assert(this!=NULL);
    bits.count = 0;
    bits.head = 0; 
    bits.malloced = 0; bits.key_mal = 1;
    int sz = bits.tablesize = size;
    if( sz > 0 ) { 
      bits.malloced = TRUE;
      PROMNEW( table, sz, T ); assert( table != NULL );
      MEMZERO( table, sz, T );
    }
    else {
      bits.malloced = FALSE; 
    }
    return 0;
  }
  int SetTable( T *tab, int sz ) {
    assert(this!=NULL); assert(bits.malloced==FALSE);
    table = tab; bits.tablesize = sz;
    MEMZERO( table, sz, T ); assert( table != NULL );
    return 0;
  }
  int SetKeyTable( int *aa, int sz ) {
    assert(this!=NULL); assert(sz==bits.tablesize);
    keytable = aa; bits.key_mal = 0; // this not malloced
    return 0 ;
  }
  // Destruction
  virtual ~PromTable(){
    assert(this!=NULL);
    if( bits.malloced ) {
      assert( table != NULL );  DELETE( table ); table = NULL;
    }
    if(keytable != NULL && bits.key_mal==1) DELETE(keytable); keytable = NULL;
  } 
  int  getCount() const { assert(this!=NULL); return bits.count; }
  int  getTableSize() const { assert(this!=NULL); return bits.tablesize; }
  bool isEmpty() const { assert(this!=NULL); return (bits.count == 0); }
  bool hasKeys()const{ return (keytable != NULL); }
  void removeAll(){ 
    assert(this!=NULL);
    bits.head = 0;  
    if( bits.count ){ 
      bits.count = 0;      assert(table!=NULL);
      MEMZERO( table, bits.tablesize, T ); 
    }
  }
#define PROM_DTAB_ZERO -999999.
  void removeAlld(){ 
    assert(this!=NULL); assert(sizeof(T) == sizeof(double));
    bits.head = 0;  
    if( bits.count ){ 
      bits.count = 0; assert(table!=NULL);
      const int lim=bits.tablesize;
      for(int ii=0;ii<lim;ii++) table[ii] = PROM_DTAB_ZERO;
    }
  }
  // destruction iteration
  T removeHead( int *pkey = NULL ){
    assert(this!=NULL); assert( bits.count ); // maintain semantics of CObList
    bits.count--;
    // get first valid at start
    int head = bits.head;
    if ( head == 0 ) {
      for( ; table[head] == (T)0 ; head++  ) /* VOID */;
      assert( head < (int)bits.tablesize );
    }
    
    // save return value
    T pOb = table[head];// return value
    table[head] = (T)0; // clear for latter use
    if( keytable ) { if(pkey) *pkey = keytable[head]; keytable[head] = 0; }
    else assert ( !pkey );
    head++;
    
    // get next valid
    if ( bits.count ) {
      for( ; table[head] == (T)0 ; head++  )
	assert( head < (int)bits.tablesize );
      bits.head = head; // new next elem
    }
    else {
      bits.head = 0; // list is empty - reset it 
    }
    return pOb;
  }
  //
  T getHead()const{
    assert(this!=NULL); assert(bits.head == 0);
    if ( !bits.count ) return (T)0;
    // find first valid place
    int i = 0;
    do {
      if ( table[i] != (T)0 ) return table[i]; 
    }while( ++i < (int)bits.tablesize );  
    PetscPrintf(MPI_COMM_SELF,"[?]%s ERROR: failed to find\n",__FUNCT__); 
    abort();  
    return (T)0;
  }
  // iteration
  TablePos getHeadPosition() const {
    assert(this!=NULL); assert( bits.head == 0 ); assert(table != NULL);
    if( bits.count == 0 ) return 0;
    
    // find first valid place
    int i = 0;
    do {
      if ( table[i] != (T)0 ) return i+1; 
    }while( ++i < (int)bits.tablesize );  
    PetscPrintf(MPI_COMM_SELF,"[?]%s ERROR: failed to find\n",__FUNCT__); 
    abort(); 
    return 0;
  }
  // TablePos is one based index into table
  T getNext( TablePos &rPosition, int *pkey = NULL ) const{
#if defined(PETSC_USE_BOPT_g)
    assert(this!=NULL); assert( bits.head == 0 ); // COblist semantics
    assert(rPosition > 0); assert(rPosition <= bits.tablesize); 
#endif
    T pOb = table[rPosition-1]; // object to return
    if( pkey ){ 
      int index = rPosition - 1; 
      *pkey = keytable[index]; 
    }
    // get next
    int i = 0;
    do{
      rPosition++;
      if( rPosition > bits.tablesize ) {      
	rPosition = 0; // end of list
	return pOb;
      }
      else if( table[rPosition-1] != (T)0 ) {  
	return pOb;
      }
    }while( ++i < (int)bits.tablesize ); 
    PetscPrintf(MPI_COMM_SELF,"[?]%s failed to find\n",__FUNCT__); abort();
    assert(0);
    return (T)0;
  }
  // add an element if it is not here -- hack for doubles
  int addd(double Elem, int key ){
#if defined(PETSC_USE_BOPT_g)
    assert(this!=NULL && bits.head==0 && Elem != PROM_DTAB_ZERO&&table!=NULL);
#endif
    if( keytable==NULL ){
      PROMNEW( keytable, bits.tablesize, int );
      int lim = bits.tablesize; // hack to intit. table w/  'PROM_DTAB_ZERO'
      for(int ii=0;ii<lim;ii++) table[ii] = PROM_DTAB_ZERO;
      bits.key_mal = 1;   
    }
    const int  tsize = bits.tablesize, tcount = bits.count; // old sizes
    int i = 0; unsigned long hash = HASH( key );
    do{
      if( table[hash] == Elem && keytable[hash] == key ){
	return 0; // < val, key > allready here
      }
      else if( table[hash] != PROM_DTAB_ZERO && keytable[hash] == key ) {
	table[hash] = Elem; // allow overright
	return -key;        // negative old value return -- hack!!!
      }
      else if ( table[hash] == PROM_DTAB_ZERO ) { // found a place 
	table[hash] = Elem; // add it
	keytable[hash] = key;
	bits.count++;
	// check for full - make sure we are efficient
	const int limit = prom_min( (tsize*3)/4, tsize - 3 );
	if( tcount >= limit ) {
	  double *oldtab = table; int *oldkt = keytable;
	  // alloc new (bigger) table
	  const int newsize = 3*tsize-1;
#if defined(PETSC_USE_BOPT_g)
	  int mype; MPI_Comm_rank(PETSC_COMM_WORLD, &mype );
	  PetscPrintf(PETSC_COMM_SELF,"[%d]D+%d->%d+",mype,tsize,newsize); 
#endif
	  bits.tablesize = newsize; // needed for HASH and HASHINC
	  bits.count = 0; 
	  PROMNEW( table, newsize, double ); 
	  for(int ii=0;ii<newsize;ii++) table[ii] = PROM_DTAB_ZERO;
	  PROMNEW( keytable, newsize, int ); bits.key_mal = 1;
	  rehashd( tsize, oldtab, oldkt );
	  // clean up
	  if ( bits.malloced ) { 
	    DELETE( oldtab );  
	    DELETE( oldkt ); 
	  }
	  bits.malloced = 1;  
	}
	else if ( tcount >= tsize - 2 ) {
	  PetscPrintf( MPI_COMM_SELF, "ERROR: hash table is full, (%d/%d)\n",
		       tcount, tsize);
	  SETERRQ(2,"ERROR -- and may not be caught!!!");
	}
	return 1;
      }
      // keep looking
      hash = HASHINC(hash);
      // }while( 1 );
    }while( i++ < tsize );
    return 1;
  }
  // add an element if it is not here
  int add(T Elem, const int inkey = -9999 ){
#if defined(PETSC_USE_BOPT_g)
    assert(Elem != (T)0);assert(this!=NULL && bits.head==0 && table != NULL);
#endif
    if( inkey != -9999 && keytable==NULL ) {
#if defined(PETSC_USE_BOPT_g)
      assert(!bits.count && bits.tablesize > 0);
#endif 
      PROMNEW( keytable, bits.tablesize, int );
      bits.key_mal = 1;
    }
    const int  tsize = bits.tablesize, tcount = bits.count; // old sizes
    int        key = (keytable!=NULL) ? inkey :  PINT(Elem);
    int i = 0; unsigned long hash = HASH( key );
    do{
      if( table[hash] == Elem && (keytable==NULL || keytable[hash]==key) ){
	return 0; // < val, key > allready here
      }
      else if( table[hash] != (T)0 && keytable && keytable[hash] == key ) {
	table[hash] = Elem; // allow overright
	return -key;        // negative old value return -- hack!!!
      }
      else if ( table[hash] == (T)0 ) { // found a place 
	table[hash] = Elem; // add it
	if( keytable ) keytable[hash] = key;
	bits.count++;
	// check for full - make sure we are efficient
	const int limit = prom_min( (tsize*2)/3, tsize - 3 );
	if( tcount >= limit ) {
	  T *oldtab = table; int *oldkt = keytable;
	  // alloc new (bigger) table
	  const int newsize = 2*tsize+1;
#if defined(PETSC_USE_BOPT_g) 
	  //int mype; MPI_Comm_rank(PETSC_COMM_WORLD, &mype );
	  //PetscPrintf(PETSC_COMM_SELF,"[%d]+%d->%d+",mype,tsize,newsize); 
#endif
	  bits.tablesize = newsize; // needed for HASH and HASHINC
	  bits.count = 0; 
	  PROMNEW( table, newsize, T ); 
	  MEMZERO( table, newsize, T );
	  if( oldkt ) { PROMNEW( keytable, newsize, int );  }
	  rehash( tsize, oldtab, oldkt );
	  // clean up
	  if( bits.malloced ) { 
	    DELETE( oldtab );  
	    if( oldkt ) DELETE( oldkt ); 
	  }  
	  bits.malloced = 1;  
	}
	else if ( tcount >= tsize - 2 ) {
	  PetscPrintf( MPI_COMM_SELF, "ERROR: hash table is full, (%d/%d)\n",
		       tcount, tsize);
	}
	return 1;
      }
      // keep looking
      hash = HASHINC(hash);
      // }while( 1 );
    }while( i++ < tsize );

    return 1;
  }
protected:
  inline void rehash( const int tsize, T *oldtab, int *oldkt ) {
    // rehash 
    for ( int ii = 0 ; ii < tsize ; ii++ ) {
      T pOb = oldtab[ii];
      if( pOb ) {
	int key = keytable ? oldkt[ii] : PINT(pOb);
	unsigned long hash = HASH( key ); // rehash - cant do an int key w/o old keys
	while (1){                // add
	  if ( table[hash] == (T)0 ) {
	    table[hash] = pOb;
	    if ( keytable ) keytable[hash] = key;
	    bits.count++;         // got it
	    break;
	  }
	  hash = HASHINC(hash); // keep looking
	}
      }
    }
  }
  inline void rehashd( const int tsize, double *oldtab, int *oldkt ) {
    // rehash 
    for ( int ii = 0 ; ii < tsize ; ii++ ) {
      assert(keytable != NULL);
      double pOb = oldtab[ii];
      if( pOb != PROM_DTAB_ZERO ) {
	int key = oldkt[ii];
	unsigned long hash = HASH( key ); // rehash - cant do an int key w/o old keys
	while (1){                // add
	  if ( table[hash] == PROM_DTAB_ZERO ) {
	    table[hash] = pOb;
	    keytable[hash] = key;
	    bits.count++;         // got it
	    break;
	  }
	  hash = HASHINC(hash); // keep looking
	}
      }
    }
  }
public:
  T find( const int key )const{
    assert( this!=NULL && bits.head == 0 && key > 0); 
    if( bits.count == 0 ) return (T)0;
    assert(keytable != NULL);
    unsigned long hash = HASH( key );
    do {
      if( table[hash] == (T)0 ) return (T)0;
      else if( keytable[hash] == key ) return table[hash];
      hash = HASHINC(hash);
    }while(1);
  }
  // T == int -- hack to get table to work with and w/o keys w/ int data
  T findi( const int Elem ) const {
    assert(keytable == NULL); // use find instead!!!
    assert( this!=NULL && bits.head == 0 ); 
    unsigned long hash = HASH( PINT(Elem) ); // PINT required for DEC!?
    do {
      if( table[hash] == (T)0 ) return (T)0;
      else if( table[hash] == Elem ) return table[hash];
      hash = HASHINC(hash);
    }while(1);
  }
  // T == double (has keys)
  double findd( const int key )const{
    assert( this!=NULL && key > 0); 
    if( bits.count == 0 ) return PROM_DTAB_ZERO;
    assert(keytable != NULL);
    unsigned long hash = HASH( key );
    const double *const tab = table;
    const int *const ktab = keytable;
    do {
      if( tab[hash] == PROM_DTAB_ZERO ) return PROM_DTAB_ZERO;
      else if( ktab[hash] == key ) return tab[hash];
      hash = HASHINC(hash);
    }while(1);
  }
  T* findp( const int key ) const {
    assert( this!=NULL && bits.head == 0 && keytable != NULL); assert(key>0); 
    if( bits.count == 0 ) return 0;
    //if(!keytable) return 0;
    unsigned long hash = HASH( key );
    do {
      if( table[hash] == (T)0 ) return 0;
      else if( keytable[hash] == key ) return &table[hash];
      hash = HASHINC(hash);
    }while(1);
  }
  // T == AthenaNode*
  T find( const AthenaNode *Elem ) const { 
    assert(this!=NULL && table!=NULL); assert(keytable == NULL);
    if( bits.count == 0 ) return (T)0;
    //assert(bits.head==0 && bits.count<bits.tablesize && keytable==NULL && Elem!=NULL);
    unsigned long hash = HASH( PINT(Elem) );
    do {
      if( table[hash] == (T)0 ) return (T)0;
      else if( (void*)table[hash] == (void*)Elem ) return table[hash];
      hash = HASHINC(hash);
    }while(1);
  }
  // T == PromDTet*
  T find( const PromDTet *Elem ) const { 
    assert(this!=NULL && table!=NULL); assert(keytable == NULL);
    if( bits.count == 0 ) return (T)0;
    //assert(bits.head==0 && bits.count<bits.tablesize && keytable==NULL && Elem!=NULL);
    int key = PINT(Elem); unsigned long hash = HASH( key );
    do {
      if( table[hash] == (T)0 ) return (T)0;
      else if( (void*)table[hash] == (void*)Elem ) return table[hash];
      hash = HASHINC(hash);
    }while(1);
  } 
  // T == PromNode*
  T find( const PromNode *Elem ) const { 
    assert(this!=NULL && table!=NULL); assert(keytable == NULL);
    if( bits.count == 0 ) return (T)0;
    //assert(bits.head == 0 && bits.count < bits.tablesize && !keytable && Elem!=NULL);
    unsigned long hash = HASH( PINT(Elem) );
    do {
      if( table[hash] == (T)0 ) return (T)0;
      else if( (void*)table[hash] == (void*)Elem ) return table[hash];
      hash = HASHINC( hash );
    }while(1);
  }
  //
  bool   isHeadZero()const{ return (bits.head == 0); }
  //
  int    isIerr()const{
    if( !this || !table ) return 1;
    int i,j;
    for(i=0,j=0;i<bits.tablesize;i++)if(table[i])j++;
    return !(j == bits.count);
  }
private:
  int         *keytable;
  T           *table;
  struct { // 96 bits 
    mutable unsigned int head : 31;
    unsigned int malloced     : 1;
    int count                 : 31;
    unsigned int key_mal      : 1;
    int tablesize             : 32;
  }bits;
};

#endif //__TABLE_H__
